@@ -1,12 +1,15 @@
 package Template::Perlish;
 
+# vim: ts=3 sts=3 sw=3 et ai :
+
 use 5.008_000;
 use warnings;
 use strict;
 use Carp;
 use English qw( -no_match_vars );
 use constant ERROR_CONTEXT => 3;
-{ our $VERSION = '1.50'; }
+{ our $VERSION = '1.52'; }
+use Scalar::Util qw< blessed reftype >;
 
 # Function-oriented interface
 sub import {
@@ -37,18 +40,53 @@ sub render {
 } ## end sub render
 
 # Object-oriented interface
-sub new {
-   my ($package, @rest) = @_;
-   my $self = bless {
-      start     => '[%',
-      stop      => '%]',
-      utf8      => 1,
-      variables => {},
-     },
-     $package;
-   %{$self} = (%{$self}, @rest == 1 ? %{$rest[0]} : @rest);
-   return $self;
-} ## end sub new
+{
+   my %preset_for;
+   BEGIN {
+      %preset_for = (
+         'default' => {
+            method_over_key => 0,
+            start  => '[%',
+            stdout => 1,
+            stop   => '%]',
+            strict_blessed => 0,
+            traverse_methods => 0,
+            utf8   => 1,
+         },
+         '1.52' => {
+            method_over_key => 1,
+            stdout => 0,
+            traverse_methods => 1,
+         },
+      );
+   }
+   sub new {
+      my $package = shift;
+      my $self = bless {%{$preset_for{'default'}}, variables => {}},
+        $package;
+
+      if (@_ == 1) {
+         %{$self} = (%{$self}, %{$_[0]});
+      }
+      elsif (scalar(@_) % 2 == 0) {
+         while (@_) {
+            my ($key, $value) = splice @_, 0, 2;
+            if ($key eq '-preset') {
+               croak "invalid preset $value in new()"
+                 unless exists $preset_for{$value};
+               %{$self} = (%{$self}, %{$preset_for{$value}});
+            }
+            else {
+               $self->{$key} = $value;
+            }
+         }
+      }
+      else {
+         croak 'invalid number of input arguments for constructor';
+      }
+      return $self;
+   } ## end sub new
+}
 
 sub process {
    my ($self, $template, $vars) = @_;
@@ -82,7 +120,7 @@ sub _compile_code_text {
 
    my $compiled = "# line 1 'input'\n";
    $compiled .= "use utf8;\n\n" if $self->{utf8};
-   $compiled .= "print {*STDOUT} '';\n\n";
+   $compiled .= "P('');\n\n";
    my $pos     = 0;
    my $line_no = 1;
    while ($pos < length $template) {
@@ -119,7 +157,7 @@ sub _compile_code_text {
             $code =~ m{\A\s* (\$ [[:alpha:]_]\w*) \s*\z}mxs)
          {
             $compiled .=
-              "\nprint {*STDOUT} $scalar; ### straight scalar\n\n";
+              "\nP($scalar); ### straight scalar\n\n";
          } ## end elsif (my ($scalar) = $code...)
          elsif (substr($code, 0, 1) eq q<=>) {
             $compiled .= "\n# line $line_no 'template<3,$line_no>'\n"
@@ -155,8 +193,14 @@ sub traverse {  ## no critic (RequireArgUnpacking,ProhibitExcessComplexity)
    my $iref         = ref($_[0]);
    my $ref_wanted   = ($iref eq 'SCALAR') || ($iref eq 'REF');
    my $ref_to_value = $ref_wanted ? shift : \shift;
-   return ($ref_wanted ? $ref_to_value : $$ref_to_value) unless @_;
 
+   # early detection of options, remove them from args list
+   my $opts = (@_ && (ref($_[-1]) eq 'HASH')) ? pop(@_) : {};
+
+   # if there's not $path provided, just don't bother going on. Actually,
+   # no $path means just return root, undefined path is always "not
+   # present" though.
+   return ($ref_wanted ? $ref_to_value : $$ref_to_value) unless @_;
    my $path_input = shift;
    return ($ref_wanted ? undef : '') unless defined $path_input;
 
@@ -172,6 +216,12 @@ sub traverse {  ## no critic (RequireArgUnpacking,ProhibitExcessComplexity)
    return ($ref_wanted ? undef : '') unless defined $crumbs;
 
    # go down the rabbit hole
+   my $use_method = $opts->{traverse_methods} || 0;
+   my ($strict_blessed, $method_pre) = (0, 0);
+   if ($use_method) {
+      $strict_blessed = $opts->{strict_blessed} || 0;
+      $method_pre = (! $strict_blessed && $opts->{method_over_key}) || 0;
+   }
    for my $crumb (@$crumbs) {
 
       # $key is what we will look into $$ref_to_value. We don't use
@@ -180,8 +230,8 @@ sub traverse {  ## no critic (RequireArgUnpacking,ProhibitExcessComplexity)
       my $key = $crumb;
 
       # $ref tells me how to look down into $$ref_to_value, i.e. as
-      # an ARRAY or a HASH
-      my $ref = ref $$ref_to_value;
+      # an ARRAY or a HASH... or object
+      my $ref = reftype $$ref_to_value;
 
       # if $ref is not true, we hit a wall. How we proceed depends on
       # whether we were asked to auto-vivify or not.
@@ -198,7 +248,7 @@ sub traverse {  ## no critic (RequireArgUnpacking,ProhibitExcessComplexity)
 
          # if $key_ref is not the same as $ref there is a mismatch
          # between what's available ($ref) and what' expected ($key_ref)
-         return $ref_wanted ? undef : '' if $key_ref ne $ref;
+         return($ref_wanted ? undef : '') if $key_ref ne $ref;
 
          # OK, data and expectations agree. Get the "real" key
          if ($key_ref eq 'ARRAY') {
@@ -216,7 +266,26 @@ sub traverse {  ## no critic (RequireArgUnpacking,ProhibitExcessComplexity)
       $ref ||= ($key =~ m{\A (?: 0 | [1-9]\d*) \z}mxs) ? 'ARRAY' : 'HASH';
 
       # time to actually do the next step
-      if ($ref eq 'HASH') {
+      my $is_blessed = blessed $$ref_to_value;
+      my $method = $is_blessed && $$ref_to_value->can($key);
+      if ($is_blessed && $strict_blessed) {
+         return($ref_wanted ? undef : '') unless $method;
+         $ref_to_value = \($$ref_to_value->$method());
+      }
+      elsif ($method && $method_pre) {
+         $ref_to_value = \($$ref_to_value->$method());
+      }
+      elsif (($ref eq 'HASH') && exists($$ref_to_value->{$key})) {
+         $ref_to_value = \($$ref_to_value->{$key});
+      }
+      elsif (($ref eq 'ARRAY') && exists($$ref_to_value->[$key])) {
+         $ref_to_value = \($$ref_to_value->[$key]);
+      }
+      elsif ($method && $use_method) {
+         $ref_to_value = \($$ref_to_value->$method());
+      }
+      # autovivification goes here eventually
+      elsif ($ref eq 'HASH') {
          $ref_to_value = \($$ref_to_value->{$key});
       }
       elsif ($ref eq 'ARRAY') {
@@ -248,72 +317,112 @@ sub _compile_sub {
    my @warnings;
    {
       my $utf8 = $self->{utf8} ? 1 : 0;
+      my $stdout = $self->{stdout} ? 1 : 0;
       local $SIG{__WARN__} = sub { push @warnings, @_ };
-      my $code = <<"END_OF_CODE";
+      my @code;
+      push @code, <<'END_OF_CODE';
    sub {
-      my \%variables = \%{\$self->{variables}};
-      my \$V = \\\%variables; # generic kid, as before by default
+      my %variables = %{$self->{variables}};
+      my $V = \%variables; # generic kid, as before by default
 
       {
-         my \$vars = shift || {};
-         if (ref(\$vars) eq 'HASH') { # old case
-            \%variables = (\%variables, \%\$vars);
+         my $vars = shift || {};
+         if (ref($vars) eq 'HASH') { # old case
+            %variables = (%variables, %$vars);
          }
          else {
-            \$V = \$vars;
-            \%variables = (HASH => { \%variables }, REF => \$V);
+            $V = $vars;
+            %variables = (HASH => { %variables }, REF => $V);
          }
       }
 
+      my $buffer = ''; # output variable
+      my $OFH;
+END_OF_CODE
+
+      my $handle = '$OFH';
+      if ($stdout) {
+         $handle = 'STDOUT';
+         push @code, <<'END_OF_CODE';
+      local *STDOUT;
+      open STDOUT, '>', \$buffer or croak "open(): $OS_ERROR";
+      $OFH = select(STDOUT);
+END_OF_CODE
+      }
+      else {
+         push @code, <<'END_OF_CODE';
+      open $OFH, '>', \$buffer or croak "open(): $OS_ERROR";
+END_OF_CODE
+      }
+
+      push @code, "binmode $handle, ':encoding(utf8)';\n"
+         if $utf8;
+
+      push @code, <<'END_OF_CODE';
+
       no warnings 'redefine';
       local *V  = sub {
-         my \$path = scalar(\@_) ? shift : [];
-         my \$input = scalar(\@_) ? shift : \$V;
-         return traverse(\$input, \$path);
+         my $path = scalar(@_) ? shift : [];
+         my $input = scalar(@_) ? shift : $V;
+         return traverse($input, $path, $self);
       };
       local *A  = sub {
-         my \$path = scalar(\@_) ? shift : [];
-         my \$input = scalar(\@_) ? shift : \$V;
-         return \@{traverse(\$input, \$path) || []};
+         my $path = scalar(@_) ? shift : [];
+         my $input = scalar(@_) ? shift : $V;
+         return @{traverse($input, $path, $self) || []};
       };
       local *H  = sub {
-         my \$path = scalar(\@_) ? shift : [];
-         my \$input = scalar(\@_) ? shift : \$V;
-         return \%{traverse(\$input, \$path) || {}};
+         my $path = scalar(@_) ? shift : [];
+         my $input = scalar(@_) ? shift : $V;
+         return %{traverse($input, $path, $self) || {}};
       };
       local *HK = sub {
-         my \$path = scalar(\@_) ? shift : [];
-         my \$input = scalar(\@_) ? shift : \$V;
-         return keys \%{traverse(\$input, \$path) || {}};
+         my $path = scalar(@_) ? shift : [];
+         my $input = scalar(@_) ? shift : $V;
+         return keys %{traverse($input, $path, $self) || {}};
       };
       local *HV = sub {
-         my \$path = scalar(\@_) ? shift : [];
-         my \$input = scalar(\@_) ? shift : \$V;
-         return values \%{traverse(\$input, \$path) || {}};
+         my $path = scalar(@_) ? shift : [];
+         my $input = scalar(@_) ? shift : $V;
+         return values %{traverse($input, $path, $self) || {}};
       };
-      use warnings 'redefine';
-
-      local *STDOUT;
-      open STDOUT, '>', \\my \$buffer or croak "open(): \$OS_ERROR";
-      binmode STDOUT, ':encoding(utf8)' if $utf8;
-      my \$previous_selection = select(STDOUT);
-      { # double closure to free "my" variables
-         my (\$buffer, \$previous_selection); # hide external ones
-         { # this enclusure allows using "my" again
-$outcome->{code_text}
-         }
-      }
-      select(\$previous_selection);
-      close STDOUT;
-      if ($utf8) {
-         require Encode;
-         \$buffer = Encode::decode(utf8 => \$buffer);
-      }
-      return \$buffer;
-   }
 END_OF_CODE
 
-      # print {*STDOUT} $code; exit 0;
+      push @code, <<"END_OF_CODE";
+      local *P = sub { return print $handle \@_; };
+      use warnings 'redefine';
+
+END_OF_CODE
+
+
+
+      push @code, <<'END_OF_CODE';
+      { # double closure to free "my" variables
+         my ($buffer, $OFH); # hide external ones
+END_OF_CODE
+
+      # the real code! one additional scope indentation to ensure we
+      # can "my" variables again
+      push @code,
+         "{\n", # this enclusure allows using "my" again
+         $outcome->{code_text},
+         "}\n}\n\n";
+
+      push @code, "select(\$OFH);\n" if $stdout;
+      push @code, "close $handle;\n\n";
+
+      if ($utf8) {
+         push @code, <<'END_OF_CODE';
+      require Encode;
+      $buffer = Encode::decode(utf8 => $buffer);
+
+END_OF_CODE
+      }
+
+      push @code, "return \$buffer;\n}\n";
+
+      my $code = join '', @code;
+      #print {*STDOUT} $code, "\n\n\n\n\n"; exit 0;
       $outcome->{sub} = eval $code;    ## no critic (ProhibitStringyEval)
       return $outcome if $outcome->{sub};
    }
@@ -355,19 +464,19 @@ sub _extract_section {
 sub _simple_text {
    my $text = shift;
 
-   return "print {*STDOUT} '$text';\n\n" if $text !~ /[\n'\\]/mxs;
+   return "P('$text');\n\n" if $text !~ /[\n'\\]/mxs;
 
    $text =~ s/^/ /gmxs;    # indent, trick taken from diff -u
    return <<"END_OF_CHUNK";
 ### Verbatim text
-print {*STDOUT} do {
+P(do {
    my \$text = <<'END_OF_INDENTED_TEXT';
 $text
 END_OF_INDENTED_TEXT
    \$text =~ s/^ //gms;      # de-indent
    substr \$text, -1, 1, ''; # get rid of added newline
    \$text;
-};
+});
 
 END_OF_CHUNK
 } ## end sub _simple_text
@@ -432,7 +541,7 @@ sub _variable {
 
    return <<"END_OF_CHUNK";
 ### Variable from the stash (\$V)
-print {*STDOUT} V([$path]);
+P(V([$path]));
 
 END_OF_CHUNK
 } ## end sub _variable
@@ -445,7 +554,7 @@ sub _expression {
    my \$value = do {{
 $expression
    }};
-   print {*STDOUT} \$value if defined \$value;
+   P(\$value) if defined \$value;
 }
 
 END_OF_CHUNK
